/*
 * Copyright (C) 2011 by egnite GmbH
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*
 * $Id$
 */

#include <dev/board.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>

#include "fc_flasher.h"

int FlashReadDesc(FLASH_DESC *desc)
{
    uint32_t i;

    outr(EEFC_FCR, EEFC_FCMD_GETD | EEFC_FKEY);
    while ((inr(EEFC_FSR) & EEFC_FRDY) == 0);
    desc->fl_id = inr(EEFC_FRR);
    desc->fl_size = inr(EEFC_FRR);
    desc->fl_page_size = inr(EEFC_FRR);
    desc->fl_nb_plane = inr(EEFC_FRR);
    desc->fl_plane = malloc(desc->fl_nb_plane * sizeof(uint32_t));
    for (i = 0; i < desc->fl_nb_plane; i++) {
        desc->fl_plane[i] = inr(EEFC_FRR);
    }
    desc->fl_nb_lock = inr(EEFC_FRR);
    desc->fl_lock = malloc(desc->fl_nb_lock * sizeof(uint32_t));
    for (i = 0; i < desc->fl_nb_lock; i++) {
        desc->fl_lock[i] = inr(EEFC_FRR);
    }
    return 0;
}

static int FlashPageWrite(uint16_t page, uint32_t *data, int len)
{
    int i;
    uint32_t addr = FLASH_BASE + page * 512;

    for (i = 0; i < len; i++) {
        outr(addr, *data);
        data++;
        addr += 4;
    }
    outr(EEFC_FCR, EEFC_FKEY | (page << EEFC_FARG_LSB) | EEFC_FCMD_EWP);
    while ((inr(EEFC_FSR) & EEFC_FRDY) == 0);

    return 0;
}

static int FlashPageRead(uint16_t page, uint32_t *data, int len)
{
    int i;
    uint32_t addr = FLASH_BASE + page * 512;

    for (i = 0; i < len; i++) {
        *data = inr(addr);
        data++;
        addr += 4;
    }
    return 0;
}

void FlashGpvnmWrite(int bit)
{
    outr(EEFC_FCR, EEFC_FKEY | (bit << EEFC_FARG_LSB) | EEFC_FCMD_SGPB);
    while ((inr(EEFC_FSR) & EEFC_FRDY) == 0);
}   

int FlashGpvnmRead(int bit)
{
    outr(EEFC_FCR, EEFC_FKEY | EEFC_FCMD_GGPB);
    while ((inr(EEFC_FSR) & EEFC_FRDY) == 0);

    return (inr(EEFC_FRR) & _BV(bit)) != 0;
}

int FlashControllerFlasher(char *image, uint32_t start, size_t pgsize)
{
    int fd;
    int got;
    uint32_t page;
    void *wbuf;
    int wcnt;
    void *rbuf;
    int rcnt;

    printf("Image '%s': ", image);
    if ((fd = _open(image, _O_RDONLY | _O_BINARY)) == -1) {
        printf("Not available, result=%d.\n", errno);
        return 0;
    }

    page = start / pgsize;
    wbuf = malloc(pgsize);
    rbuf = malloc(pgsize);
    if (wbuf == NULL || rbuf == NULL) {
        printf("Out of memory. Really?\n");
        return -1;
    }

    printf("Starting at %08lX, page %lu\n", start + FLASH_BASE, page);
    wcnt = 0;
    rcnt = 0;
    for (;; page++) {
        if ((got = _read(fd, wbuf, pgsize)) != 0) {
            if (got < 0) {
                printf("Error %d: Reading image file.\n", errno);
                free(wbuf);
                free(rbuf);
                return -1;
            }
            if (got) {
                rcnt += got;
                if (FlashPageRead(page, rbuf, (got + 3) / 4)) {
                    printf("Error: Flash read failed at page %lu.\n", page);
                    free(wbuf);
                    free(rbuf);
                    return -1;
                }
                if (memcmp(wbuf, rbuf, got)) {
                    if (got < pgsize) {
                        memset((uint8_t *)wbuf + got, 0xFF, pgsize - got);
                    }
                    if (FlashPageWrite(page, wbuf, pgsize / 4)) {
                        printf("Error: Flash write failed at page %lu.\n", page);
                        free(wbuf);
                        free(rbuf);
                        return -1;
                    }
                    if (FlashPageRead(page, rbuf, pgsize / 4)) {
                        printf("Error: Flash read failed at page %lu.\n", page);
                        free(wbuf);
                        free(rbuf);
                        return -1;
                    }
                    if (memcmp(wbuf, rbuf, pgsize)) {
                        printf("Error: Flash compare failed at page %lu.\n", page);
                        free(wbuf);
                        free(rbuf);
                        return -1;
                    }
                    wcnt += pgsize;
                }
            }
        }
        if (got < pgsize) {
            break;
        }
    }
    _close(fd);
    free(wbuf);
    free(rbuf);
    printf("%d of %d bytes updated.\n", wcnt, rcnt);

    return 0;
}

